# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

class MachineType(object):
    none = 1
    motor = 2
    hinge = 3
    spring = 4
    light = 5
    prismatic_joint = 6

